/*
 *   This program is intended to invoke wget
 *   on any URL you paste into it.
 *
 *   wget will be invoked recursively with
 *   recursion-depth of one and with "Camino"
 *   as the Agent-String.
 *
 *   The Agent-String is needed for some webpages who
 *   disallow recursive downloads with wget, and only
 *   allow downloads, when there is a User-Agent's name.
 *
 *   The main intention here is: download recursively
 *   a webpage (with depth == 1), which's URL is
 *   thrown in here.
 *
 *   This reduces a lot of handy work, when finding URLs
 *   you want to download.
 *
 *   Without this program you would have to call wget for every of those URLs
 *   or at least you would have to write the URLs to a file and then invoke
 *   wget on using this URL-file.  This would be much more work than only
 *   pasting in the URL into this program.
 *
*/


#include <stdio.h>
#include <stdlib.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <unistd.h>
#include <fcntl.h>

#include <sys/time.h>
#include <time.h>

#include <signal.h>

#include<pthread.h>

#include<curses.h>
#include<ncurses.h>


/* -------------------------------------------------------------- */
/* -------------------------------------------------------------- */
/* -------------------------------------------------------------- */



#define KEY_WAIT_MS 100 /* 100 ms waiting-time */

#define STRING_SIZE_MAX 10000
char raw_last_entered_url[ STRING_SIZE_MAX + 1 ];
char raw_buffer[ STRING_SIZE_MAX + 1 ];
int  raw_buf_idx = 0;


int glob_num = 0;

int raw_num_of_childs    = 0;
int raw_num_of_downloads = 0;


/* CLI-options / defaults */
/* ====================== */
char* myopts = "u:d:vh"; /* accepted options */
typedef struct {
                 char*   user_agent_name;           /* -u <user-agent-name> */
                 short   depth;                     /* -d <num_depth>       */
                 short   print_version;             /* -v                   */
                 short   print_help_message;        /* -h                   */
               } CLI_OPTIONS;

CLI_OPTIONS cli_options = { "Camino",
                             2,
                             0,
                             0
                          }; /* Default's setzen */




pthread_mutex_t raw_buf_mutex    = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t raw_screen_mutex = PTHREAD_MUTEX_INITIALIZER;




FILE* url_log   = NULL;
FILE* main_log  = NULL;
//FILE* err_log   = NULL;


char* main_log_file = "invwget.log";
char* url_log_file  = "wget-urls.txt";
//char* err_log_file  = "invwget_err.log";




struct timeval raw_this_time;
struct timeval raw_last_keystroke_time;
struct timezone raw_tz;

int raw_time_diff    = 0;
int raw_time_diff_ms = 0;



struct _data_ {
  char* last_entered_url;
  char* buffer;
  int  buf_idx;
  pthread_mutex_t* buf_mutex_p;

  int  num_of_childs;
  int  num_of_downloads;
  struct timeval this_time;
  struct timeval last_keystroke_time;
  struct timezone tz;
  int time_diff;
  int time_diff_ms;
} data;




/* -------------------------- */
/* -------------------------- */
/* -------------------------- */
void writelog(FILE* log,  char* msg )
{
  fprintf(log, "%s\n", msg);
  fflush(log);
  return;
}

/* -------------------------------------------------------- */
/* -------------------------------------------------------- */
/* -------------------------------------------------------- */
void Gettimeofday( struct timeval *tv, struct timezone *tz )
{
  int ret = gettimeofday( tv, tz );
  if( ret )
  {
    writelog(main_log, "Gettimeofday: gettimeofday()-error\n");
    abort();
  }
  return;
}

/* -------------------------- */
/* -------------------------- */
/* -------------------------- */
void init_synchronization()
{
  data.last_entered_url    = raw_last_entered_url;
  data.buffer              = raw_buffer;
  data.buf_idx             = raw_buf_idx;
  data.buf_mutex_p         = &raw_buf_mutex;

  data.num_of_childs       = raw_num_of_childs;
  data.num_of_downloads    = raw_num_of_downloads;

  data.this_time           = raw_this_time;
  data.last_keystroke_time = raw_last_keystroke_time;
  data.tz                  = raw_tz;

  /* ignore the timezone */
  /* ------------------- */
  data.tz.tz_minuteswest = 0;
  data.tz.tz_dsttime     = 0;

  Gettimeofday( &data.this_time, &data.tz );
  Gettimeofday( &data.last_keystroke_time, &data.tz );

  return;
}



/* ----------------------------------- */
/* ----------------------------------- */
/* ----------------------------------- */
void Pthread_mutex_lock( pthread_mutex_t *mutex )
{
  int retval = 0;

    retval = pthread_mutex_lock( mutex );
    if( retval ) /* should never oocur! */
    {
      writelog(main_log, "enter_keys(): pthread_mutex_lock() failed\n");
      abort();
    }

  return;
}


/* ----------------------------------- */
/* ----------------------------------- */
/* ----------------------------------- */
void Pthread_mutex_unlock( pthread_mutex_t *mutex )
{
  int retval = 0;

  retval = pthread_mutex_unlock( mutex );
  if( retval ) /* should never oocur! */
    abort();

  return;
}




/* ----------------------------------- */
/* ----------------------------------- */
/* ----------------------------------- */
void shutdown_ncurses()
{
  endwin();
}






/* ----------------------------------- */
/* ----------------------------------- */
/* ----------------------------------- */
void update_screen()
{
/* maybe using colors...
//int res = 0;
//res = init_color(COLOR_BLACK, 100, 100, 100); // nur, wenn Farb-Definitionen geaendert wwerden sollen
//if( res == ERR ) flash();

// eigentliches aktivieren der (vorhandenen) Colors
init_pair(1, COLOR_RED, COLOR_BLACK);
attron(COLOR_PAIR(1));
*/

  move(0,0);
  printw("Num of Childs:    %4d\n", data.num_of_childs);
  printw("Num of Downloads: %4d\n", data.num_of_downloads);
  printw("Last entered URL: %s\n",  data.last_entered_url);

  move(5,0);
  printw("Pressing the ESC-key ends the program without cancelling the downloads\n");
  printw("Terminating via CTRL-C kills all downloading childs!\n");
  
  refresh();

  return;
}



/* -------------------------- */
/* -------------------------- */
/* -------------------------- */
void* enter_keys( void* params )
{
  int ch = 0;

  while( 1 )
  {
    ch = getch();

    /* if ESCape-key, then end! */
    /* ------------------------ */
    if( ch == 27 )
      exit(EXIT_SUCCESS); /* successful ending; childs will be inherited by init */

    /* ------------------------------------------------------ */
    /* Eingabe ist noch in Aktion => speichere neues Zeichen. */
    /* ------------------------------------------------------ */

    Pthread_mutex_lock( data.buf_mutex_p );

      if( data.buf_idx < STRING_SIZE_MAX )
      {
        data.buffer[ data.buf_idx ] = ch;
        ++data.buf_idx;

        Gettimeofday( &data.last_keystroke_time, &data.tz );
      }

    Pthread_mutex_unlock( data.buf_mutex_p );
  }
  return NULL;
}


/* -------------------- */
/* wget will be started */
/* -------------------- */
void child_code()
{
  int fd     = 0;

  fprintf(url_log, "%s\n", data.buffer);
  fflush( url_log );

  /* ------------------------------------------------------------- */
  /* wget writes to stderr, and this we don't want                 */
  /* so we open /dev/null and making STDERR_FILENO a copy of it :) */
  /* ------------------------------------------------------------- */
  fd = open("/dev/null", O_WRONLY);
  if( fd > -1 )
  {
    close( STDERR_FILENO );
    dup2( fd, STDERR_FILENO );
  }
  else
  {
    writelog(main_log, "opening /dev/null failed!\n");
    abort();
  }

    writelog(main_log, cli_options.user_agent_name);
  /* execing the wget :) */
  /* ------------------- */
  execlp( "wget", "wget", "-r", "-l", "2", "-U", cli_options.user_agent_name, data.buffer, NULL );
  /*                                  ^^^_ cli_options.depth  als string */

  /* if exec would fail... */
  /* --------------------- */
  fprintf(stderr, "Could not exec! with %s\n", data.buffer);
  perror("");
  writelog(main_log, "exec() of wget failed!\n");
  abort();
}



/* -------------------------- */
/* -------------------------- */
/* -------------------------- */
void* invoke_wget( void* params )
{
  pid_t pid  = 0;


  while(1) /* hier besser Abfragen einer Condition-variable (pthread-maessig) */
  {
    Pthread_mutex_lock( data.buf_mutex_p );

      Gettimeofday( &data.this_time, &data.tz );

      data.time_diff = (data.this_time.tv_sec - data.last_keystroke_time.tv_sec) * 1000000
                    + (data.this_time.tv_usec - data.last_keystroke_time.tv_usec);
      data.time_diff_ms = data.time_diff / 1000;

      /* --------------------------------------------------------------------- */
      /* if the last entry was long enough ago and we already have data in the */
      /* buffer then we start one wget-program.                                */
      /* --------------------------------------------------------------------- */
      if( data.time_diff_ms > KEY_WAIT_MS && data.buf_idx > 0 )
      {
        /* --------------------------- */
        /* diese Eingabe ist komplett, */
        /* also starte wget.           */
        /* --------------------------- */

        /* the URL-buffer is copied to the "last_entered_url", */
        /* which will be printed on screen.                    */
        /* --------------------------------------------------- */
        memcpy( data.last_entered_url, data.buffer, sizeof(raw_last_entered_url) );


        /* -------------------------------- */
        /* OK, let's make a copy of our own */
        /* -------------------------------- */
        pid = fork();

        /* fork() FAILED! */
        if( pid < 0 )
        {
          writelog(main_log, "invoke_wget(): fork() failed\n");
          abort();
        }

        if( pid == 0 ) /* CHILD => invoke wget */
          child_code();


        /* ------ */
        /* Parent */
        /* ------ */
        ++data.num_of_childs;
        ++data.num_of_downloads;
        bzero( data.buffer, sizeof(raw_buffer) ); /* empty buffer for new input */
        data.buf_idx = 0;
        Gettimeofday( &data.last_keystroke_time, &data.tz );
        update_screen();

      Pthread_mutex_unlock( data.buf_mutex_p );

    }
    else
    {
      Pthread_mutex_unlock( data.buf_mutex_p );
      usleep(1000*KEY_WAIT_MS); /* we wait for a while */
    }

  }

  return NULL;
}

/* ------------------------- */
/* Signalhandler for SIGCHLD */
/* ------------------------- */
/* calls wait() to get the   */
/* return-status of a child. */
/* ------------------------- */
void handle_sigchld()
{
  int   status = 0;
  pid_t pid    = 0;

  pid = wait( &status );
  if( pid != -1 )
    --data.num_of_childs;
  else
    writelog(main_log, "Wait failed!\n");

  update_screen();

  return;
}


/* ===================================================================== */

/* --------------------------------------------------------------------- */
int print_help_message( FILE* outstream )
{
  fprintf( outstream, "usage: ");
  fprintf( outstream,  "invwget [-u <agent_name>] [-v] [-h]\n");
  fprintf( outstream, " option u: User-Agent name\n");
  fprintf( outstream, " option v: print version.\n");
  fprintf( outstream, " option h: print this help message.\n");
  return 0;
}



int clipping( int minval, int val, int maxval )
{
  val = val < minval ? minval : val;
  val = val > maxval ? maxval : val;

  return val;
}

/* --------------------------------------------------------------------- */
int get_cli_options( int argc, char* argv[], CLI_OPTIONS* options )
{
  extern CLI_OPTIONS cli_options; /* the global variable for the CLI-options */
  int c;
  extern char* myopts;
  extern void  print_usage();
  extern char *optarg;

  while((c = getopt( argc, argv, myopts ) ) != -1 )
  {
printf("Optarg: %s\n", optarg); // REMOVE
    switch( c )
    {
      //case '?': print_usage(stderr); exit(EXIT_FAILURE);
      case 'u': options->user_agent_name = optarg;   break;
      case 'd': options->depth = (short) clipping( 1, atoi(optarg), 5 ); break; // HARDCODED
      case 'v': options->print_version   = 1;        break;
      case 'h': options->print_help_message = 1;     break;
    }
  }

  return 0;
}


/* ===================================================================== */
/* ===================================================================== */
/* ===================================================================== */
int main( int argc, char* argv[] )
{
  extern FILE* main_log;
  extern FILE* url_log;


  pthread_t    keys_thread;
  pthread_t    wget_thread;

  int   retval   = 0;
  struct sigaction sigactions;


  get_cli_options( argc, argv, &cli_options );


  sigactions.sa_handler = handle_sigchld;
  sigactions.sa_flags   = 0;
  sigemptyset( &sigactions.sa_mask );
  sigaction( SIGCHLD, &sigactions, NULL );

  atexit( shutdown_ncurses );


  main_log = fopen( main_log_file, "a" );
  if( main_log == NULL )
    abort();

  url_log = fopen( url_log_file, "a" );
  if( url_log == NULL )
    abort();



  /* initialize */
  /* ---------- */
  init_synchronization();



  /* NCURSES-Initialisierung */
  /* ----------------------- */
  initscr();

  cbreak(); /* YEAH! */
  noecho();
  keypad(stdscr, TRUE);

  start_color();

  update_screen();

  retval = pthread_create( &keys_thread, NULL, enter_keys, NULL );
  retval = pthread_create( &wget_thread, NULL, invoke_wget, NULL ); /* here makes sense: PTHREAD_CREATE_DETACHED */


  pthread_join( keys_thread, NULL ); /* wait here until keys_thread finishes */

  /*
  endwin(); at exit...
  */

  fclose( url_log );
  fclose( main_log );

  return 0;
}
